@@ -2,17 +2,35 @@
 AutoBOT. Skrlj et al. 2021
 """
 
+from sklearn import preprocessing
+from sklearn.preprocessing import Normalizer
+from sklearn import pipeline
+from sklearn.pipeline import FeatureUnion
+from sklearn.base import BaseEstimator, TransformerMixin
+from .topic_features import *
+from .doc_similarity import *
+from .conceptnet_features import *
+from .keyword_features import *
+from .sentence_embeddings import *
+from .word_relations import *
+import string
+import re
+from nltk import pos_tag
+import multiprocessing as mp
+from nltk.corpus import stopwords
+from nltk import word_tokenize
+import pandas as pd
+from sklearn.model_selection import GridSearchCV
+from sklearn.svm import LinearSVC
+from sklearn.linear_model import LogisticRegression, SGDClassifier
+from scipy.sparse import hstack
+from sklearn.feature_extraction.text import TfidfVectorizer
+import numpy as np
 import logging
 logging.basicConfig(format='%(asctime)s - %(message)s',
                     datefmt='%d-%b-%y %H:%M:%S')
 logging.getLogger().setLevel(logging.INFO)
 
-import numpy as np
-from sklearn.feature_extraction.text import TfidfVectorizer
-from scipy.sparse import hstack
-from sklearn.linear_model import LogisticRegression, SGDClassifier
-from sklearn.svm import LinearSVC
-from sklearn.model_selection import GridSearchCV
 
 try:
     import nltk
@@ -22,13 +40,6 @@ except Exception as es:
     import nltk
     print(es)
 
-import pandas as pd
-from nltk import word_tokenize
-from nltk.corpus import stopwords
-import multiprocessing as mp
-from nltk import pos_tag
-import re
-import string
 try:
     from nltk.tag import PerceptronTagger
 except:
@@ -36,22 +47,57 @@ except:
     def PerceptronTagger():
         return 0
 
+# Feature constructors
 
-## Feature constructors
-from .word_relations import *
-from .sentence_embeddings import *
-from .keyword_features import *
-from .conceptnet_features import *
-from .doc_similarity import *
-from .topic_features import *
+# sklearn dependencies
 
-## sklearn dependencies
-from sklearn.base import BaseEstimator, TransformerMixin
-from sklearn.pipeline import FeatureUnion
-from sklearn import pipeline
-from sklearn.preprocessing import Normalizer
-from sklearn import preprocessing
+# Import screening for contextual features
+global contextual_feature_library
 
+try:
+    from .contextual_embeddings import *
+    contextual_feature_library = True
+
+except:
+    contextual_feature_library = False
+
+global feature_presets
+feature_presets = {}
+
+# Full stack
+feature_presets['neurosymbolic'] = [
+    'concept_features', 'document_graph', 'neural_features_dbow',
+    'neural_features_dm', 'relational_features_token', 'topic_features',
+    'keyword_features', 'relational_features_char', 'char_features',
+    'word_features', 'pos_features', 'contextual_features'
+]
+
+# This one is ~language agnostic
+feature_presets['neurosymbolic-lite'] = [
+    'document_graph', 'neural_features_dbow', 'neural_features_dm',
+    'topic_features', 'keyword_features', 'relational_features_char',
+    'char_features', 'word_features'
+]
+
+# MLJ paper versions
+feature_presets['neurosymbolic-default'] = [
+    'neural_features_dbow', 'neural_features_dm', 'keyword_features',
+    'relational_features_char', 'char_features', 'word_features', "pos_features",
+    'concept_features'
+]
+
+feature_presets['neural'] = [
+    'document_graph', 'neural_features_dbow', 'neural_features_dm'
+]
+
+feature_presets['symbolic'] = [
+    'concept_features', 'relational_features_token', 'topic_features',
+    'keyword_features', 'relational_features_char', 'char_features',
+    'word_features', 'pos_features'
+]
+
+if not contextual_feature_library:
+    feature_presets['neurosymbolic'].pop()
 
 def remove_punctuation(text):
     """
@@ -80,7 +126,7 @@ def remove_stopwords(text):
 def remove_mentions(text, replace_token):
     """
     This method removes mentions (relevant for tweets)
-    
+
     :param str text: Input string of text
     :param str replace_token: A token to be replaced
     :return str string: A new text string
@@ -117,7 +163,7 @@ def remove_url(text, replace_token):
 def get_affix(text):
     """
     This method gets the affix information
-    
+
     :param str text: Input text.
     """
 
@@ -141,7 +187,7 @@ def get_pos_tags(text):
 def ttr(text):
     """
     Number of unique tokens
-    
+
     :param str text: Input string of text
     :return float floatValue: Ratio of the unique/overall tokens
     """
@@ -160,6 +206,7 @@ class text_col(BaseEstimator, TransformerMixin):
     :param obj TransformerMixin: Transformer object
     :return obj object: Returns particular text column
     """
+
     def __init__(self, key):
         self.key = key
 
@@ -170,15 +217,16 @@ class text_col(BaseEstimator, TransformerMixin):
         return data_dict[self.key]
 
 
-#fit and transform numeric features, used in scikit Feature union
+# fit and transform numeric features, used in scikit Feature union
 class digit_col(BaseEstimator, TransformerMixin):
     """
     Dealing with numeric features
-    
+
     :param obj BaseExtimator: Core estimator
     :param obj TransformerMixin: Transformer object
     :return obj object: Returns transformed (scaled) space
     """
+
     def fit(self, x, y=None):
         return self
 
@@ -233,6 +281,7 @@ class FeaturePrunner:
     """
     Core class describing sentence embedding methodology employed here.
     """
+
     def __init__(self, max_num_feat=2048):
 
         self.max_num_feat = max_num_feat
@@ -258,16 +307,16 @@ def fast_screening_sgd(training, targets):
         f1_scoring = "f1_macro"
     else:
         f1_scoring = "f1"
-    ncores = 8
-    clf = GridSearchCV(svc,
-                       parameters,
-                       verbose=0,
-                       n_jobs=ncores,
-                       cv=5,
-                       scoring=f1_scoring,
-                       refit=False)
-    clf.fit(training, targets)
-    f1_perf = max(clf.cv_results_['mean_test_score'])
+        ncores = 8
+        clf = GridSearchCV(svc,
+                           parameters,
+                           verbose=0,
+                           n_jobs=ncores,
+                           cv=5,
+                           scoring=f1_scoring,
+                           refit=False)
+        clf.fit(training, targets)
+        f1_perf = max(clf.cv_results_['mean_test_score'])
     return f1_perf
 
 
@@ -282,7 +331,7 @@ def get_subset(indice_list, data_matrix, vectorizer):
                                  current_fnum:(current_fnum +
                                                min(num_feat, len(features)))]
             feature_subspaces.append(subset)
-        current_fnum += len(features)
+            current_fnum += len(features)
     return hstack(feature_subspaces)
 
 
@@ -334,8 +383,9 @@ def get_features(df_data,
                  embedding_dim=512,
                  memory_location="memory/conceptnet.txt.gz",
                  custom_pipeline=None,
-                 concept_features=True,
+                 concept_features=False,
                  random_seed=54324,
+                 contextual_model="paraphrase-xlm-r-multilingual-v1",
                  combine_with_existing_representation=False):
     """
     Method that computes various TF-IDF-alike features.
@@ -348,12 +398,13 @@ def get_features(df_data,
     :param str memory_location: Location of the gzipped ConceptNet-like memory.
     :param obj custom_pipeline: Custom pipeline to be used for features if needed.
     :param bool concept_features: Whether to use ConceptNet features -- needs the internet connection.
+    :param str contextual_model: The language model string compatible with sentence-transformers library (this is in beta)
     :param int random_seed: The seed for the pseudo-random parts.
     :param bool combine_with_existing_representation: Whether to use existing representations + user-specified ones.
     :return obj/list/matrix: Transformer pipeline, feature names and the feature matrix.
     """
 
-    ## Seeds and np for np
+    # Seeds and np for np
     np.random.seed(random_seed)
 
     if not custom_pipeline is None and combine_with_existing_representation == False:
@@ -368,135 +419,150 @@ def get_features(df_data,
             format(max_num_feat, sparsity))
 
         max_tokenizer, max_feature_names, max_data_matrix = None, None, None
-        logging.info("Constructing {} features.".format(representation_type))
 
-        if representation_type == "neural" or "neurosymbolic" in representation_type:
+        # Initialize transformer objects
+        sentence_embedder_dm1 = documentEmbedder(max_features=max_num_feat,
+                                                 dm=1,
+                                                 ndim=embedding_dim)
 
-            sentence_embedder_dm1 = documentEmbedder(max_features=max_num_feat,
-                                                     dm=1,
-                                                     ndim=embedding_dim)
+        sentence_embedder_dm2 = documentEmbedder(max_features=max_num_feat,
+                                                 dm=0,
+                                                 ndim=embedding_dim)
 
-            sentence_embedder_dm2 = documentEmbedder(max_features=max_num_feat,
-                                                     dm=0,
-                                                     ndim=embedding_dim)
+        doc_sim_features = RelationalDocs(ndim=embedding_dim, targets=None)
 
-            doc_sim_features = RelationalDocs(ndim=embedding_dim, targets=None)
+        tfidf_word_unigram = TfidfVectorizer(ngram_range=(1, 3),
+                                             sublinear_tf=False,
+                                             max_features=max_num_feat)
 
-            neural_features = [
-                ('neural_features_dm',
-                 pipeline.Pipeline([('s7', text_col(key='no_stopwords')),
-                                    ('sentence_embedding_mean',
-                                     sentence_embedder_dm1), ('normalize', Normalizer(norm = "max"))])),
-                ('neural_features_dbow',
-                 pipeline.Pipeline([('s8', text_col(key='no_stopwords')),
-                                    ('sentence_embedding_mean',
-                                     sentence_embedder_dm2), ('normalize', Normalizer(norm = "max"))])),
-                ('document_graph',
-                 pipeline.Pipeline([('s9', text_col(key='no_stopwords')),
-                                    ('doc_similarity_features',
-                                     doc_sim_features), ('normalize', Normalizer(norm = "max"))]))
+        tfidf_pos_unigram = TfidfVectorizer(ngram_range=(1, 3),
+                                            max_features=max_num_feat)
+
+        tfidf_char_bigram = TfidfVectorizer(analyzer='char',
+                                            ngram_range=(2, 4),
+                                            max_features=max_num_feat)
+
+        lr_rel_features_unigram = relationExtractor(max_features=max_num_feat,
+                                                    min_token="unigrams")
+
+        lr_rel_features_token = relationExtractor(max_features=max_num_feat,
+                                                  min_token="word")
+
+        keyword_features = KeywordFeatures(max_features=max_num_feat,
+                                           targets=targets)
+
+        topic_features = TopicDocs(ndim=embedding_dim)
+
+        concept_features_transformer = ConceptFeatures(max_features=max_num_feat,
+                                                       knowledge_graph=memory_location)
+        contextual_features = None
+        
+        if contextual_feature_library:
+            
+            if representation_type == "neurosymbolic":
+                contextual_features = ContextualDocs(model=contextual_model)
+                
+            elif isinstance(representation_type, list) and "neurosymbolic" in representation_type:
+                contextual_features = ContextualDocs(model=contextual_model)            
+
+        feature_transformer_vault = {
+            "pos_features":
+            ('pos_features',
+             pipeline.Pipeline([('s3', text_col(key='pos_tag_seq')),
+                                ('pos_tfidf_unigram', tfidf_pos_unigram),
+                                ('normalize', Normalizer(norm="max"))])),
+            "word_features":
+            ('word_features',
+             pipeline.Pipeline([('s1', text_col(key='no_stopwords')),
+                                ('word_tfidf_unigram', tfidf_word_unigram),
+                                ('normalize', Normalizer(norm="max"))])),
+            "char_features":
+            ('char_features',
+             pipeline.Pipeline([('s2', text_col(key='no_stopwords')),
+                                ('char_tfidf_bigram', tfidf_char_bigram),
+                                ('normalize', Normalizer(norm="max"))])),
+            "relational_features_char":
+            ('relational_features_char',
+             pipeline.Pipeline([('s4', text_col(key='no_stopwords')),
+                                ('relational_features_unigram',
+                                 lr_rel_features_unigram),
+                                ('normalize', Normalizer(norm="max"))])),
+            "keyword_features": ('keyword_features',
+                                 pipeline.Pipeline([
+                                     ('s5', text_col(key='no_stopwords')),
+                                     ('keyword_features', keyword_features),
+                                     ('normalize', Normalizer(norm="max"))
+                                 ])),
+            "topic_features": ('topic_features',
+                               pipeline.Pipeline([
+                                   ('s6', text_col(key='no_stopwords')),
+                                   ('topic_features', topic_features),
+                                   ('normalize', Normalizer(norm="max"))
+                               ])),
+            "relational_features_token":
+            ('relational_features_token',
+             pipeline.Pipeline([('s4', text_col(key='no_stopwords')),
+                                ('relational_features_token',
+                                 lr_rel_features_token),
+                                ('normalize', Normalizer(norm="max"))])),
+            "neural_features_dm": ('neural_features_dm',
+                                   pipeline.Pipeline([
+                                       ('s7', text_col(key='no_stopwords')),
+                                       ('sentence_embedding_mean',
+                                        sentence_embedder_dm1),
+                                       ('normalize', Normalizer(norm="max"))
+                                   ])),
+            "neural_features_dbow": ('neural_features_dbow',
+                                     pipeline.Pipeline([
+                                         ('s8', text_col(key='no_stopwords')),
+                                         ('sentence_embedding_mean',
+                                          sentence_embedder_dm2),
+                                         ('normalize', Normalizer(norm="max"))
+                                     ])),
+            "document_graph":
+            ('document_graph',
+             pipeline.Pipeline([('s9', text_col(key='no_stopwords')),
+                                ('doc_similarity_features', doc_sim_features),
+                                ('normalize', Normalizer(norm="max"))])),
+            "concept_features": ('concept_features',
+                                 pipeline.Pipeline([
+                                     ('s6', text_col(key='no_stopwords')),
+                                     ('concept_features',
+                                      concept_features_transformer),
+                                     ('normalize', Normalizer(norm="max"))
+                                 ])),
+            "contextual_features": ('contextual_features',
+                                    pipeline.Pipeline([
+                                        ('s6', text_col(key='text')),
+                                        ('concept_features', contextual_features),
+                                        ('normalize', Normalizer(norm="max"))
+                                    ]))
+        }
+
+        if isinstance(representation_type, str):
+
+            ## representation_type is pre-set
+
+            if not representation_type in feature_presets:
+                logging.info(
+                    "Please, specify a valid preset! (see the documentation for the up-to-date namings)")
+
+            preset_features = feature_presets[representation_type]
+            if not concept_features:
+                preset_features = [
+                    x for x in preset_features if x != "concept_features"]
+
+            features = [
+                feature_transformer_vault[x]
+                for x in preset_features
             ]
 
-            if "neurosymbolic-default" in representation_type or "neurosymbolic-lite" in representation_type:
-                neural_features.pop()
-
-        if representation_type == "symbolic" or "neurosymbolic" in representation_type:
-
-            tfidf_word_unigram = TfidfVectorizer(ngram_range=(1, 3),
-                                                 sublinear_tf=False,
-                                                 max_features=max_num_feat)
-
-            tfidf_pos_unigram = TfidfVectorizer(ngram_range=(1, 3),
-                                                max_features=max_num_feat)
-
-            tfidf_char_bigram = TfidfVectorizer(analyzer='char',
-                                                ngram_range=(2, 4),
-                                                max_features=max_num_feat)
-
-            lr_rel_features_unigram = relationExtractor(
-                max_features=max_num_feat, min_token="unigrams")
-
-            lr_rel_features_token = relationExtractor(
-                max_features=max_num_feat, min_token="word")
-
-            keyword_features = KeywordFeatures(max_features=max_num_feat,
-                                               targets=targets)
-
-            topic_features = TopicDocs(ndim=embedding_dim)
-
-            symbolic_features = [
-                ('pos_features',
-                 pipeline.Pipeline([('s3', text_col(key='pos_tag_seq')),
-                                    ('pos_tfidf_unigram', tfidf_pos_unigram),
-                                    ('normalize', Normalizer(norm = "max"))
-                                    ])),
-                ('word_features',
-                 pipeline.Pipeline([('s1', text_col(key='no_stopwords')),
-                                    ('word_tfidf_unigram', tfidf_word_unigram),
-                                    ('normalize', Normalizer(norm = "max"))
-                                    ])),
-                ('char_features',
-                 pipeline.Pipeline([('s2', text_col(key='no_stopwords')),
-                                    ('char_tfidf_bigram', tfidf_char_bigram),
-                                    ('normalize', Normalizer(norm = "max"))
-                                    ])),
-                ('relational_features_char',
-                 pipeline.Pipeline([('s4', text_col(key='no_stopwords')),
-                                    ('relational_features_unigram',
-                                     lr_rel_features_unigram),
-                                    ('normalize', Normalizer(norm = "max"))])),
-                ('keyword_features',
-                 pipeline.Pipeline([('s5', text_col(key='no_stopwords')),
-                                    ('keyword_features', keyword_features),
-                                    ('normalize', Normalizer(norm = "max"))])),
-                ('topic_features',
-                 pipeline.Pipeline([('s6', text_col(key='no_stopwords')),
-                                    ('topic_features', topic_features),
-                                    ('normalize', Normalizer(norm = "max"))])),
-                ('relational_features_token',
-                 pipeline.Pipeline([('s4', text_col(key='no_stopwords')),
-                                    ('relational_features_token',
-                                     lr_rel_features_token),
-                                    ('normalize', Normalizer(norm = "max"))])),
-            ]
-
-            if "neurosymbolic-default" in representation_type:
-
-                ## Last two feature types are new since the initial publication.
-                symbolic_features.pop()
-                symbolic_features.pop()
-
-            elif representation_type == "neurosymbolic-lite":
-
-                ## Remove the document graph-based features (beta)
-                symbolic_features.pop()
-                symbolic_features = symbolic_features[1:] ## Remove POS-tags
-
-            elif representation_type == "neurosymbolic":
-
-                ## Remove POS tags (not multilingual)
-                symbolic_features = symbolic_features[1:]
-
-            if concept_features:
-                concept_features = ConceptFeatures(
-                    max_features=max_num_feat, knowledge_graph=memory_location)
-
-                cfx = ('concept_features',
-                       pipeline.Pipeline([('s6', text_col(key='no_stopwords')),
-                                          ('concept_features',
-                                           concept_features),
-                                          ('normalize', Normalizer(norm = "max"))]))
-
-                symbolic_features.append(cfx)
-
-        if representation_type == "symbolic":
-            features = symbolic_features
-
-        elif "neurosymbolic" in representation_type:
-            features = symbolic_features + neural_features
-
         else:
-            features = neural_features
+
+            # representation_type is a list of desired feature types.
+            features = [
+                feature_transformer_vault[x] for x in representation_type
+            ]
 
     if not custom_pipeline is None and combine_with_existing_representation:
         features = features + custom_pipeline
@@ -527,6 +593,7 @@ def get_autoBOT_manual(train_sequences,
                        max_features=1000,
                        clf_type="LR"):
 
+    # This is legacy now.
     total_sequences_training = train_sequences.values.tolist(
     ) + dev_sequences.values.tolist()
     total_sequences_training = build_dataframe(total_sequences_training)
@@ -567,7 +634,8 @@ def get_autoBOT_manual(train_sequences,
                                                     min_token="unigrams")
         keyword_features = KeywordFeatures(max_features=max_num_feat,
                                            targets=train_targets)
-        concept_features = ConceptFeatures(max_features=max_num_feat)
+        concept_features_transformer = ConceptFeatures(
+            max_features=max_num_feat)
         symbolic_features = [
             ('word_features',
              pipeline.Pipeline([('s1', text_col(key='no_stopwords')),
@@ -587,7 +655,7 @@ def get_autoBOT_manual(train_sequences,
                                 ('keyword_features', keyword_features)])),
             ('concept_features',
              pipeline.Pipeline([('s6', text_col(key='no_stopwords')),
-                                ('concept_features', concept_features)]))
+                                ('concept_features', concept_features_transformer)]))
         ]
 
     features = symbolic_features + neural_features
